from abc import ABC, abstractmethod
from typing import List, Optional
from lsprotocol.types import CompletionList, CompletionParams, Diagnostic, Position
from pygls.server import LanguageServer
from pygls.workspace import Document


class Analyzer(ABC):
    @abstractmethod
    def hover_info(self, doc: Document, pos: Position) -> Optional[str]:
        pass

    @abstractmethod
    def get_diagnostics(self, doc: Document) -> List[Diagnostic]:
        pass

    @abstractmethod
    def get_completions(
        self, ls: LanguageServer, params: CompletionParams
    ) -> CompletionList:
        pass



import threading


class Debouncer:
    def __init__(self, wait):
        self.wait = wait
        self.timer = None
        self.lock = threading.Lock()

    def debounce(self, func):
        def debounced(*args, **kwargs):
            with self.lock:
                if self.timer is not None:
                    self.timer.cancel()  # Cancel the existing timer if there is one
                # Create a new timer that will call func with the latest arguments
                self.timer = threading.Timer(self.wait, lambda: func(*args, **kwargs))
                self.timer.start()

        return debounced



import re
from typing import List, Optional
from lark import UnexpectedInput, UnexpectedToken
from packaging.specifiers import Specifier
from packaging.version import Version
from lsprotocol.types import (
    CompletionList,
    CompletionParams,
    Diagnostic,
    Position,
    Range,
)
from pygls.server import LanguageServer
from pygls.workspace import Document
from vvm.exceptions import VyperError
from vyper_lsp.analyzer.BaseAnalyzer import Analyzer
import vvm

from pathlib import Path

from lark import Lark
from vyper.ast.grammar import PythonIndenter

GRAMMAR_FILE_PATH = Path(__file__).parent.parent / "grammar" / "grammar.lark"
GRAMMAR = GRAMMAR_FILE_PATH.read_text()

parser = Lark(GRAMMAR, parser="lalr", start="module", postlex=PythonIndenter())


def format_parse_error(e):
    if isinstance(e, UnexpectedToken):
        expected = ", ".join(e.accepts or e.expected)
        return f"Unexpected token '{e.token}' at {e.line}:{e.column}. Expected one of: {expected}"
    else:
        return str(e)


LEGACY_VERSION_PRAGMA_REGEX = re.compile(r"^#\s*@version\s+(.*)$")
VERSION_PRAGMA_REGEX = re.compile(r"^#pragma\s+version\s+(.*)$")


def extract_version_pragma(line: str) -> Optional[str]:
    if match := LEGACY_VERSION_PRAGMA_REGEX.match(line):
        return match.group(1)
    elif match := VERSION_PRAGMA_REGEX.match(line):
        return match.group(1)
    else:
        return None


# regex that matches numbers and underscores
# ex: 1_000_000
NUMBER_REGEX = re.compile(r"^(_\d+)*$")


class SourceAnalyzer(Analyzer):
    def __init__(self) -> None:
        self.parser_enabled = True
        self.compiler_enabled = False

    def get_version_pragma(self, doc: Document) -> Optional[str]:
        doc_lines = doc.lines
        for line in doc_lines:
            if version := extract_version_pragma(line):
                return version

    def hover_info(self, doc: Document, pos: Position) -> Optional[str]:
        return None

    def get_parser_diagnostics(self, doc: Document) -> List[Diagnostic]:
        diagnostics = []
        last_error = None

        def on_grammar_error(e: UnexpectedInput) -> bool:
            nonlocal last_error
            if (
                last_error is not None
                and last_error.line == e.line
                and last_error.column == e.column
            ):
                return True
            if (
                last_error is not None
                and type(last_error) == type(e)
                and last_error.line == e.line
            ):
                return True
            diagnostics.append(
                Diagnostic(
                    range=Range(
                        start=Position(line=e.line - 1, character=e.column - 1),
                        end=Position(line=e.line - 1, character=e.column),
                    ),
                    message=format_parse_error(e),
                )
            )
            last_error = e
            return True

        try:
            parser.parse(doc.source, on_error=on_grammar_error)
        except Exception:
            # ignore errors that are already handled by on_error
            pass

        return diagnostics

    def get_compiler_diagnostics(self, doc: Document) -> List[Diagnostic]:
        diagnostics = []

        # now compile via vvm to get semantic errors
        try:
            vvm.compile_source(doc.source)
        except VyperError as e:
            if "vyper.exceptions.VersionException" in e.stderr_data:
                # ignore version errors, we handle them separately
                version_pragma = self.get_version_pragma(doc)
                # check if version pragma is just a version or a specifier
                if version_pragma is None:
                    diagnostics.append(
                        Diagnostic(
                            range=Range(
                                start=Position(line=0, character=0),
                                end=Position(line=0, character=0),
                            ),
                            message=str(e),
                        )
                    )
                elif re.match(r"^\d+\.\d+\.\d+$", version_pragma):
                    # version pragma is just a version, so we just install it
                    version: Version = Version(version_pragma)
                    vvm.install_vyper(version)
                    vvm.set_vyper_version(version)
                else:
                    specifier: Specifier = Specifier(version_pragma)
                    for version in vvm.get_installable_vyper_versions():
                        if specifier.contains(version):
                            vvm.install_vyper(version)
                            vvm.set_vyper_version(version)
            else:
                # find the line and number of the error
                # will appear in the text as `line x:y`
                regex = re.compile(r"line (\d+):(\d+)")
                match = regex.search(e.stderr_data)

                type_regex = re.compile(r"vyper\.exceptions\.([a-zA-Z]+): (.*)\n")
                type_match = type_regex.search(e.stderr_data)
                if match and type_match:
                    diagnostics.append(
                        Diagnostic(
                            range=Range(
                                start=Position(
                                    line=int(match.group(1)) - 1,
                                    character=int(match.group(2)) - 1,
                                ),
                                end=Position(
                                    line=int(match.group(1)) - 1,
                                    character=int(match.group(2)),
                                ),
                            ),
                            message=f"{type_match.group(1)}: {type_match.group(2)}",
                        )
                    )
        except Exception:
            pass

        return diagnostics

    def get_diagnostics(self, doc: Document) -> List[Diagnostic]:
        diagnostics = []
        if self.parser_enabled:
            diagnostics.extend(self.get_parser_diagnostics(doc))
        if self.compiler_enabled:
            diagnostics.extend(self.get_compiler_diagnostics(doc))
        return diagnostics

    def get_completions(
        self, ls: LanguageServer, params: CompletionParams
    ) -> CompletionList:
        completions: CompletionList = CompletionList(is_incomplete=False, items=[])
        return completions



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['vyper_lsp', 'vyper_lsp.analyzer', 'vyper_lsp.grammar']

package_data = \
{'': ['*']}

install_requires = \
['lark>=1.1.7,<2.0.0',
 'loguru>=0.6.0,<0.7.0',
 'lsprotocol>=2023.0.0b1,<2024.0.0',
 'packaging>=23.1,<24.0',
 'pydantic>=1.10,<2.0',
 'pygls>=1.1.2,<2.0.0',
 'tree-sitter>=0.20.1,<0.21.0',
 'vvm>=0.2.0,<0.3.0',
 'vyper>=0.3.7,<0.4.0']

entry_points = \
{'console_scripts': ['vyper-lsp = vyper_lsp.main:main']}

setup_kwargs = {
    'name': 'vyper-lsp',
    'version': '0.0.7',
    'description': 'Language server for Vyper, a pythonic smart contract language',
    'long_description': '# Vyper LSP Server\n\n## Requirements\n\nVyper LSP requires a minimum vyper version of 0.3.7. For full support, it is also required that the Vyper version installed in your virtual environment is capable of compiling your contract.\n\nFor example, a vyper contract with `#pragma version 0.3.8` cannot be compiled with `0.3.10`, so you must install `vyper==0.3.8` if you want full support while working with this contract.\n\nA contract with `#pragma version >=0.3.8` will work fine with any installed vyper version greater than the requirement, so you can get full support while editing this contract if you have the latest vyper version installed.\n\n## Install Vyper-LSP\n\n### via `pipx`\nI like `pipx` because it handles creating an isolated env for executables and putting them on your path.\n\n`pipx install git+https://github.com/vyperlang/vyper-lsp.git`\n\n### via `pip`\nYou can install using `pip` if you manage your environments through some other means:\n\n> TODO: publish on pypi\n\n`pip install git+https://github.com/vyperlang/vyper-lsp.git`\n\n## Verify installation\n\nCheck that `vyper-lsp` is on your path:\n\nIn your terminal, run `which vyper-lsp`. If installation was succesful, you should see the path to your installed executable.\n\n## Editor Setup\n\n### Emacs\n\nThe following emacs lisp snippet will create a Vyper mode derived from Python Mode, and sets up vyper-lsp.\n\n``` emacs-lisp\n(define-derived-mode vyper-mode python-mode "Vyper" "Major mode for editing Vyper.")\n\n(add-to-list \'auto-mode-alist \'("\\\\.vy\\\\\'" . vyper-mode))\n\n(with-eval-after-load \'lsp-mode\n  (add-to-list \'lsp-language-id-configuration\n               \'(vyper-mode . "vyper"))\n  (lsp-register-client\n   (make-lsp-client :new-connection\n                    (lsp-stdio-connection `(,(executable-find "vyper-lsp")))\n                    :activation-fn (lsp-activate-on "vyper")\n                    :server-id \'vyper-lsp)))\n```\n\n### Neovim\n\nAdd the following to your `neovim` lua config.\n\nIt should be at `~/.config/nvim/init.lua`\n\n``` lua\nvim.api.nvim_create_autocmd({ "BufEnter" }, {\n  pattern = { "*.vy" },\n  callback = function()\n    vim.lsp.start({\n      name = "vyper-lsp",\n      cmd = { "vyper-lsp" },\n      root_dir = vim.fs.dirname(vim.fs.find({ ".git" }, { upward = true })[1])\n    })\n  end,\n})\n\nvim.api.nvim_create_autocmd({ "BufEnter" }, {\n  pattern = { "*.vy" },\n  callback = function()\n    vim.lsp.start({\n      name = "vyper-lsp",\n      cmd = { "vyper-lsp" },\n      root_dir = vim.fs.dirname(vim.fs.find({ ".git" }, { upward = true })[1])\n    })\n  end,\n})\n\nvim.api.nvim_set_keymap(\'n\', \'gd\', \'<Cmd>lua vim.lsp.buf.definition()<CR>\', { noremap = true, silent = true })\nvim.api.nvim_set_keymap(\'n\', \'gD\', \'<Cmd>lua vim.lsp.buf.declaration()<CR>\', { noremap = true, silent = true })\nvim.api.nvim_set_keymap(\'n\', \'gr\', \'<Cmd>lua vim.lsp.buf.references()<CR>\', { noremap = true, silent = true })\nvim.api.nvim_set_keymap(\'n\', \'gi\', \'<Cmd>lua vim.lsp.buf.implementation()<CR>\', { noremap = true, silent = true })\nvim.api.nvim_set_keymap(\'n\', \'K\', \'<Cmd>lua vim.lsp.buf.hover()<CR>\', { noremap = true, silent = true })\nvim.api.nvim_set_keymap(\'n\', \'<C-k>\', \'<Cmd>lua vim.lsp.buf.signature_help()<CR>\', { noremap = true, silent = true })\nvim.api.nvim_set_keymap(\'n\', \'[d\', \'<Cmd>lua vim.lsp.diagnostic.goto_prev()<CR>\', { noremap = true, silent = true })\nvim.api.nvim_set_keymap(\'n\', \']d\', \'<Cmd>lua vim.lsp.diagnostic.goto_next()<CR>\', { noremap = true, silent = true })\n\n```\n\n\n### VS Code\n\nSee `vyper-lsp` VS Code extension\n',
    'author': 'z80',
    'author_email': 'z80@ophy.xyz',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<3.12',
}


setup(**setup_kwargs)
